# ZTOOLS
# Developer tools for use with VtM: Bloodlines
# version 0.3 ALPHA - 22 Apr 2008
# by ZylonBane
# Map entity reading/writing routines based on VPKTool, by Turvy
# 
# To install:
# 1. Copy this file to Vampire\python\
# 2. Add the line "from ztools import *" to vamputil.py
# 3. Delete vamputil.pyc and ztools.pyc (if present)

import __main__
from struct import *
Find = __main__.FindEntityByName
FindList = __main__.FindEntitiesByName
FindClass = __main__.FindEntitiesByClass

# init persistent variables
pzEntList = []
pzGrabbed = []
pzMarkers = {}
pzTempID = 0
pzPlayerTransformed = 0
pzOldPlayerModel = ""
pzOldPlayerOrg = ()
pzOldPlayerAng = ()

# dump entity data for current map
def zdumpmap():
	# structs
	header_struct = "4sl"
	lump_struct = "ll"

	# filename stuff
	dumpname = "Vampire/maps/" + getMapName("txt")
	mapname = "Vampire/maps/" + getMapName()
	bsp = open(mapname, "rb")

	# grab header
	bsp_data = bsp.read(calcsize(header_struct))
	bsp_ident, bsp_ver = unpack(header_struct, bsp_data)
	if (bsp_ident != "VBSP"):
		print "zERR: Not a Source engine BSP file"
	else:
		# grab Entities lump (lump zero)
		bsp_data = bsp.read(calcsize(lump_struct))
		lump_offset, lump_len = unpack(lump_struct, bsp_data)
		# dump lump
		dump = open(dumpname, "w")
		bsp.seek(lump_offset)
		dump.write(bsp.read(lump_len - 1))
		dump.close()
		print "MAP DUMP COMPLETE"
		print "Written to \"" + dumpname + "\""
	bsp.close()

# update entity data for current map
def zupdatemap():
	# constants
	magic_word = "QnDbTm"
	linefeed = chr(10)
	null_terminator = chr(0)

	# structs
	header_struct = "4sl"
	lump_struct = "ll"
	mod_offset_struct = "l"

	# open input file
	mapname = "Vampire/maps/" + getMapName()
	dumpname = "Vampire/maps/" + getMapName("txt")
	try:
		dump = open(dumpname, "rb")
	except:
		print "zERR: Could not open input file \"" + dumpname + "\""
		return

	# grab input data
	dump_raw = dump.read()
	dump.close()
	line_end = determineTerminator(dump_raw)
	if line_end == "":
		print "zERR: No valid data found in input file"
		return
	dump_lines = dump_raw.split(line_end)
	# clean up trailing newline
	if dump_lines[-1] == "":
		dump_lines.pop()

	# open BSP file
	bsp = open(mapname, "rb+")

	# grab header
	bsp_data = bsp.read(calcsize(header_struct))
	bsp_ident, bsp_ver = unpack(header_struct, bsp_data)
	if (bsp_ident != "VBSP"):
		print "zERR: Not a Source engine BSP file"
	else:
		# check if map previously modified
		bsp.seek(-len(magic_word), 2)
		bsp_data = bsp.read(len(magic_word))
		if bsp_data == magic_word:
			print "Found previous modding information. Replacing."
			# grab previous offset
			bsp.seek(-10, 2)
			bsp_data = bsp.read(calcsize(mod_offset_struct))
			mod_offset = unpack(mod_offset_struct, bsp_data)[0]
			# remove old mod info
			bsp.seek(mod_offset)
			bsp.truncate()
		# write new mod info
		brace_depth = 0
		line_num = 0
		bsp.seek(0, 2)
		mod_offset = bsp.tell()
		for line in dump_lines:
			line_num += 1
			bsp.write(line.rstrip() + linefeed)
			# sanity check braces
			if line[0:1] == "{":
				brace_depth += 1
			elif line[-1:1] == "}":
				brace_depth -= 1
			if brace_depth < 0:
				print "zERR: Unexpected closing brace at line #" + str(line_num)
				brace_depth = 0
			elif brace_depth > 1:
				print "zERR: Unexpected opening brace at line #" + str(line_num)
				brace_depth = 1
		# cap off mod info
		bsp.write(null_terminator + pack(mod_offset_struct, mod_offset) + magic_word)
		# write new lump header
		mod_size = bsp.tell() - mod_offset
		bsp.seek(8)
		bsp_data = pack(lump_struct, mod_offset, mod_size)
		bsp.write(bsp_data)		
	bsp.close()
	print "MAP UPDATE COMPLETE"
	print "Updated \"" + mapname + "\""

# dump content of G
def zdumpg():
	G = __main__.G
	dumpfile = "ztools_G_dump.txt"
	f = open(dumpfile, "w")
	f.write("Dump of __main__.G global persistent database.\n")
	f.write("Generated from within map \"" + getMapName() + "\", at Story State " + str(G.Story_State) +".\n\n")
	keyList = G.keys()
	keyList.sort()
	for k in keyList:
		f.write(k + " = " + str(G[k]) + "\n")
	# prettyprint morgue dictionary
	f.write("\nG.morgue:\n")
	f.write("========================================\n")
	morgueKeys = G.morgue.keys()
	morgueKeys.sort()
	for k in morgueKeys:
		f.write("   " + k + "\n")
	f.write("\nDUMP COMPLETE\n")
	f.close()
	print "DUMP COMPLETE"
	print "Written to \"" + dumpfile + "\""

# open all taxi destinations
def ztaxi():
	G = __main__.G
	flagList = (
		"Chinatown_Open",
		"Downtown_Open",
		"Giovanni_Open",
		"Hollywood_Open",
		"Kingsway_Open",
		"Mansion_Open",
		"Museum_Open",
		"Society_Open"
	)
	for flag in flagList:
		G[flag] = 1
	print "ALL TAXI DESTINATIONS OPEN"
	print "Now you just need to find a taxi."

# spawn arbitrary object into world
# optionally accepts modelname
# this doesn't work particularly well; may be removed
def zspawn(*args):
	entClass = args[0]
	if (len(args) > 1):
		entMdl = args[1]
	else:
		entMdl = ""
	pc = __main__.FindPlayer()
	x, y, z = pc.GetCenter()
	# add slight offset so new entity doesn't get false detected as being in player inventory
	z += 0.0001
	ent = __main__.CreateEntityNoSpawn(entClass, (x, y, z), (0, 0, 0))
	if entMdl != "":
		ent.SetModel(entMdl)
	__main__.CallEntitySpawn(ent)
	print "ENTITY SPAWNED AT PLAYER CENTER"

# ======================================================================
# NPC functions
# ======================================================================

# kill all NPCs
def zkill():
	npcList = FindClass("npc*")
	for npc in npcList:
		if npc.GetName() != "cabbie":
			npc.Kill()
	print "ALL NPCS KILLED"

# make all NPCs hate you
def zhate():
	npcList = FindClass("npc*")
	for npc in npcList:
		if npc.GetName() != "cabbie":
			npc.SetRelationship("player D_HT 10")
	print "ALL NPCS HATE YOU"

# make all NPCs love you
def zlove():
	npcList = FindClass("npc*")
	for npc in npcList:
		if npc.GetName() != "cabbie":
			npc.SetRelationship("player D_LI 10")
	print "ALL NPCS LOVE YOU"

# ======================================================================
# grabee origin-setting functions
# ======================================================================

# become current grabee (copy model and coordinates to player)
def zbe():
	global pzPlayerTransformed, pzOldPlayerModel, pzOldPlayerOrg
	if noGrabee(): return
	pc = __main__.FindPlayer()
	ent = pzGrabbed[0]
	entModel = ent.GetModelName()
	entOrg = ent.GetOrigin()
	entAng = ent.GetAngles()
	# decide what to do
	if entOrg == (0, 0, 0):
		print "zERR: Cannot become this entity"
	elif entModel == "":
		print "zERR: Grabee has no physical model"
	elif pzPlayerTransformed:
		# already transformed, switch back
		zunbe()
		print "UN-TRANSFORMED"
	else:
		# transform!
		pzPlayerTransformed = 1
		pzOldPlayerModel = pc.GetModelName()
		pzOldPlayerOrg = pc.GetOrigin()
		pzOldPlayerAng = pc.GetAngles()
		pc.SetModel(entModel)
		pc.SetOrigin(entOrg)
		pc.SetAngles(entAng)
		ent.ScriptHide()
		print "TRANSFORMED"
		print "Switch into third-person view if you haven't already done so."

# revert transformation
def zunbe():
	global pzPlayerTransformed, pzOldPlayerModel, pzOldPlayerOrg
	# sanity checks
	if noGrabee(): return
	if pzPlayerTransformed == 0: return
	# swap everything back
	pc = __main__.FindPlayer()
	ent = pzGrabbed[0]
	pzPlayerTransformed = 0
	ent.SetOrigin(pc.GetOrigin())
	ent.SetAngles(pc.GetAngles())
	ent.ScriptUnhide()
	pc.SetModel(pzOldPlayerModel)
	pc.SetOrigin(pzOldPlayerOrg)
	pc.SetAngles(pzOldPlayerAng)

# center grabee on player
def zhere():
	if noGrabee(): return
	pc = __main__.FindPlayer()
	x, y, z = pc.GetCenter()
	# add slight offset so new entity doesn't get false detected as being in player inventory
	z += 0.0001
	ent = pzGrabbed[0]
	ent.SetOrigin((x, y, z))
	ent.SetAngles(pc.GetAngles())
	print "ORIGIN SET TO PLAYER CENTER"
	print "You may need to 'noclip' just about now."

# center player on grabee
def zthere():
	if noGrabee(): return
	pc = __main__.FindPlayer()
	ent = pzGrabbed[0]
	pc.SetOrigin(ent.GetCenter())
	print "PLAYER CENTERED ON GRABEE ORIGIN"
	print "You may need to 'noclip' just about now."

# set grabee origin
def zorg(*args):
	if argsInvalid(args, 3): return
	ent = pzGrabbed[0]
	ent.SetOrigin((args[0], args[1], args[2]))
	print "ORIGIN SET"

# set grabee X origin
def zx(*args):
	if argsInvalid(args, 1): return
	ent = pzGrabbed[0]
	x, y, z = ent.GetOrigin()
	x = args[0]
	ent.SetOrigin((x, y, z))
	print "X ORIGIN SET TO " + str(x)

# increment grabee X origin
def zix(*args):
	if argsInvalid(args, 1): return
	ent = pzGrabbed[0]
	x, y, z = ent.GetOrigin()
	x += args[0]
	ent.SetOrigin((x, y, z))
	print "X ORIGIN INCREMENTED TO " + str(x)

# set grabee Y origin
def zy(*args):
	if argsInvalid(args, 1): return
	ent = pzGrabbed[0]
	x, y, z = ent.GetOrigin()
	y = args[0]
	ent.SetOrigin((x, y, z))
	print "Y ORIGIN SET TO " + str(y)

# increment grabee Y origin
def ziy(*args):
	if argsInvalid(args, 1): return
	ent = pzGrabbed[0]
	x, y, z = ent.GetOrigin()
	y += args[0]
	ent.SetOrigin((x, y, z))
	print "Y ORIGIN INCREMENTED TO " + str(y)

# set grabee Z origin
def zz(*args):
	if argsInvalid(args, 1): return
	ent = pzGrabbed[0]
	x, y, z = ent.GetOrigin()
	z = args[0]
	ent.SetOrigin((x, y, z))
	print "Z ORIGIN SET TO " + str(z)

# increment grabee Z origin
def ziz(*args):
	if argsInvalid(args, 1): return
	ent = pzGrabbed[0]
	x, y, z = ent.GetOrigin()
	z += args[0]
	ent.SetOrigin((x, y, z))
	print "Z ORIGIN INCREMENTED TO " + str(z)

# ======================================================================
# grabee angle-setting functions
# ======================================================================

# set grabee angles
def zang(*args):
	if argsInvalid(args, 3): return
	ent = pzGrabbed[0]
	ent.SetAngles((args[0], args[1], args[2]))
	print "ANGLES SET"

# set grabee roll angle
def zr(*args):
	if argsInvalid(args, 1): return
	ent = pzGrabbed[0]
	r, w, p = ent.GetAngles()
	r = args[0]
	ent.SetAngles((r, w, p))
	print "ROLL ANGLE SET TO " + str(args[0])

# increment grabee roll angle
def zir(*args):
	if argsInvalid(args, 1): return
	ent = pzGrabbed[0]
	r, w, p = ent.GetAngles()
	newAng = r + args[0]
	if newAng < 0: newAng += 360
	if newAng > 360: newAng -= 360
	ent.SetAngles((newAng, w, p))
	print "ROLL ANGLE INCREMENTED TO " + str(args[0])

# set grabee yaw angle
def zw(*args):
	if argsInvalid(args, 1): return
	ent = pzGrabbed[0]
	r, w, p = ent.GetAngles()
	w = args[0]
	ent.SetAngles((r, w, p))
	print "YAW ANGLE SET TO " + str(args[0])

# increment grabee yaw angle
def ziw(*args):
	if argsInvalid(args, 1): return
	ent = pzGrabbed[0]
	r, w, p = ent.GetAngles()
	newAng = w + args[0]
	if newAng < 0: newAng += 360
	if newAng > 360: newAng -= 360
	ent.SetAngles((r, newAng, p))
	print "YAW ANGLE INCREMENTED TO " + str(args[0])

# set grabee pitch angle
def zp(*args):
	if argsInvalid(args, 1): return
	ent = pzGrabbed[0]
	r, w, p = ent.GetAngles()
	p = args[0]
	ent.SetAngles((r, w, p))
	print "PITCH ANGLE SET TO " + str(args[0])

# increment grabee pitch angle
def zip(*args):
	if argsInvalid(args, 1): return
	ent = pzGrabbed[0]
	r, w, p = ent.GetAngles()
	newAng = p + args[0]
	if newAng < 0: newAng += 360
	if newAng > 360: newAng -= 360
	ent.SetAngles((r, w, newAng))
	print "PITCH ANGLE INCREMENTED TO " + str(args[0])

# ======================================================================
# visibility functions
# ======================================================================

# hide grabee
def zhide(*args):
	if len(args) == 1:
		# hide all class members
		entList = FindClass(args[0])
		for ent in entList:
			ent.ScriptHide()
	else:
		# hide grabee
		if noGrabee(): return
		pzGrabbed[0].ScriptHide()
	print "ENTITY HIDDEN"

# unhide grabee
def zunhide(*args):
	if len(args) == 1:
		# hide all class members
		entList = FindClass(args[0])
		for ent in entList:
			ent.ScriptUnhide()
	else:
		# hide grabee
		if noGrabee(): return
		pzGrabbed[0].ScriptUnhide()
	print "ENTITY UNHIDDEN"

# mark things with inspection nodes
def zmark(*args):
	if len(args) == 1:
		c = 0
		# mark all class members
		entList = FindClass(args[0])
		for ent in entList:
			if zmark_util(ent):
				c += 1
		print str(c) + " ENTITIES OF CLASS \"" + args[0] + "\" MARKED"
	else:
		# mark grabee
		if noGrabee(): return
		ent = pzGrabbed[0]
		if zmark_util(ent):
			print "ENTITY MARKED"
		else:
			print "ENTITY ALREADY MARKED"
	
# utility function for zmark
def zmark_util(ent):
	global pzMarkers
	global pzTempID
	pc = __main__.FindPlayer()
	if pc.GetOrigin() == ent.GetOrigin():
		# inventory entity
		return 0
	if not pzMarkers.has_key(ent):
		entName = ent.GetName()
		tmpName = "_ztools_tempname"
		# must assign temporary unique targetname in case of entities with no name,
		# or multiple entities with same name
		ent.SetName(tmpName)
		marker = __main__.CreateEntityNoSpawn("inspection_node", ent.GetCenter(), (0, 0, 0))
		marker.SetParent(tmpName)
		__main__.CallEntitySpawn(marker)
		# restore original name
		ent.SetName(entName)
		# add inspection entity pointer to persistent list, so we don't have to
		# screw around with temporary targetnames when removing them
		pzMarkers[ent] = marker
		return 1
	else:
		return 0

# unmark grabee
def zunmark(*args):
	if len(args) == 1:
		c = 0
		# unmark all class members
		entList = FindClass(args[0])
		for ent in entList:
			if zunmark_util(ent):
				c += 1
		print str(c) + " ENTITIES OF CLASS \"" + args[0] + "\" UNMARKED"
	else:		
		# unmark grabee
		if noGrabee(): return
		ent = pzGrabbed[0]
		if zunmark_util(ent):
			print "ENTITY UNMARKED"
		else:
			print "ENTITY NOT MARKED"

# utility function for zunmark
def zunmark_util(ent):
	global pzMarkers
	if not pzMarkers.has_key(ent):
		return 0
	else:
		marker = pzMarkers[ent]
		marker.Kill()
		del pzMarkers[ent]
		return 1

# ======================================================================
# other grabee functions
# ======================================================================

# grab entity
def zgrab(*args):
	global pzGrabbed
	if len(args) == 0:
		if pzGrabbed == []:
			print "zERR: No entity currently grabbed. zgrab() usage is:"
			print "   zgrab() - Display currently grabbed entity."
			print "   zgrab(#) - Grab entity by number from znear() list."
			print "   zgrab(\"targetname\") - Grab entity by map targetname."
		else:
			print "Currently grabbed entity is:"
			entDump(pzGrabbed[0], pzGrabbed[1])
	elif type(args[0]) == type("string"):
		# grab by targetname
		if pzGrabbed != []: zunbe()
		ent = Find(args[0])
		if ent == "":
			print "zERR: Could not find an entity with that targetname."
		else:
			pzGrabbed = (ent, "[class unknown]")
			print "ENTITY GRABBED:"
			entDump(pzGrabbed[0], pzGrabbed[1])
	elif type(args[0]) == type(1):
		# grab by entity list index
		if pzGrabbed != []: zunbe()
		i = args[0] - 1
		if (i < 0) or (i > len(pzEntList)):
			print "zERR: Not that many entities in znear() list."
		else:
			pzGrabbed = (pzEntList[i][0], pzEntList[i][1])
			print "ENTITY GRABBED:"
			entDump(pzGrabbed[0], pzGrabbed[1])
	else:
		# wrong data type passed
		print "zERR: Invalid data type. This function accepts entity list numbers or targetnames (in quotes!) only."

# zgrab shortcut
def zg(*args):
	if len(args) == 1:
		zgrab(args[0])
	else:
		zgrab()

# display info on currently grabbed entity
def zinf():
	if noGrabee(): return
	print "GRABBED ENTITY:"
	entDump(pzGrabbed[0], pzGrabbed[1])

# ======================================================================
# entity indexing functions
# ======================================================================

# display currently stored nearby list
def zlist():
	global pzEntList
	print "CURRENT ZNEAR LIST"
	if pzEntList == []:
		print "Entity list is empty. Use znear() to populate it."
		return
	i = 1;
	for entData in pzEntList:
		ent = entData[0]
		entClass = entData[1]
		entName = ent.GetName()
		entCent = ent.GetCenter()
		entMdl = ent.GetModelName()
		if (entName == ""):
			entName = "[no targetname]"
		if (entMdl == ""):
			entMdl = "[no model]"
		else:
			entMdl = entMdl[entMdl.rfind("/") + 1:]
		print str(i) + ". " + entClass + ", " + entName + ", " + entMdl + ", \"" + strH(entCent[0]) + " " + strH(entCent[1]) + " " + strH(entCent[2]) + "\""
		i = i + 1
	return

# display entities near to the player
def znear(*args):
	global pzEntList
	pc = __main__.FindPlayer()
	maxItems = 15	# max items to list
	# snag entities of interest
	entClassList = (
		"ambient_generic",
		"ambient_soundscheme",
		"beam",
		"camera_animated",
		"camera_cinematic",
		"camera_keyframe",
		"camera_track",
		"dynamic_prop",
		"entityflame",
		"env_beam",
		"env_dustpuff",
		"env_fade",
		"env_fire",
		"env_firesensor",
		"env_firesource",
		"env_floating_camera",
		"env_fog_controller",
		"env_global",
		"env_glow",
		"env_laser",
		"env_message",
		"env_microphone",
		"env_particle",
		"env_particle_hud",
		"env_physexplosion",
		"env_physimpact",
		"env_player_surface_trigger",
		"env_shake",
		"env_shooter",
		"env_soundscape",
		"env_spark",
		"env_sprite",
		"env_steam",
		"env_steamjet",
		"env_sun",
		"env_swarm_bats",
		"env_terrainmorph",
		"env_wind",
		"flame_cluster",
		"func_areaportal",
		"func_areaportalwindow",
		"func_breakable",
		"func_breakable_surf",
		"func_brush",
		"func_button",
		"func_door",
		"func_door_rotating",
		"func_dustcloud",
		"func_dustmotes",
		"func_elevator",
		"func_illusionary",
		"func_keyframed_mover",
		"func_lod",
		"func_monitor",
		"func_movelinear",
		"func_particle",
		"func_physbox",
		"func_plat",
		"func_platrot",
		"func_pushable",
		"func_rot_button",
		"func_rotating",
		"func_wall",
		"game_sign",
		"game_text",
		"gib",
		"gibshooter",
		"info_landmark",
		"info_target",
		"info_teleport_destination",
		"infodecal",
		"inspection_node",
		"inspection_brush",
		"intersting_place",
		"intersting_place_conversation",
		"item_a_body_armor",
		"item_a_hvy_cloth",
		"item_a_hvy_leather",
		"item_a_lt_cloth",
		"item_a_lt_leather",
		"item_container",
		"item_container_animated",
		"item_container_lock",
		"item_container_one_item_filtered",
		"item_d_holy_light",
		"item_g_astrolite",
		"item_g_bach_journal",
		"item_g_badlucktalisman",
		"item_g_bailbond_receipt",
		"item_g_bertrams_cd",
		"item_g_bloodpack",
		"item_g_bluebloodpack",
		"item_g_brotherhood_flyer",
		"item_g_car_stereo",
		"item_g_cash_box",
		"item_g_driver_license_gimble",
		"item_g_drugs_drug_box",
		"item_g_drugs_morphine_bottle",
		"item_g_drugs_perscription_bottle",
		"item_g_drugs_pill_bottle",
		"item_g_edane_print_report",
		"item_g_edane_report",
		"item_g_eldervitaepack",
		"item_g_eyes",
		"item_g_gargoyle_book",
		"item_g_garys_cd",
		"item_g_garys_film",
		"item_g_garys_photo",
		"item_g_garys_tape",
		"item_g_ghost_pendant",
		"item_g_giovanni_invitation_maria",
		"item_g_giovanni_invitation_victor",
		"item_g_guy_magazine",
		"item_g_hannahs_appt_book",
		"item_g_hatters_screenplay",
		"item_g_horrortape_1",
		"item_g_horrortape_2",
		"item_g_idol_cat",
		"item_g_idol_crane",
		"item_g_idol_dragon",
		"item_g_idol_elephant",
		"item_g_jumbles_flyer",
		"item_g_junkyard_businesscard",
		"item_g_keyring",
		"item_g_larry_briefcase",
		"item_g_lilly_diary",
		"item_g_lilly_photo",
		"item_g_lilly_purse",
		"item_g_lockpick",
		"item_g_mercurio_journal",
		"item_g_milligans_businesscard",
		"item_g_oh_diary",
		"item_g_pisha_book",
		"item_g_pisha_fetish",
		"item_g_ring03",
		"item_g_ring_gold",
		"item_g_ring_serial_killer_1",
		"item_g_ring_silver",
		"item_g_sewerbook_1",
		"item_g_stake",
		"item_g_vampyr_apocrypha",
		"item_g_vv_photo",
		"item_g_wallet",
		"item_g_warr_clipboard",
		"item_g_warr_ledger_1",
		"item_g_warr_ledger_2",
		"item_g_warrens4_passkey",
		"item_g_watch_fancy",
		"item_g_watch_normal",
		"item_g_werewolf_bloodpack",
		"item_g_wireless_camera_1",
		"item_g_wireless_camera_2",
		"item_g_wireless_camera_3",
		"item_g_wireless_camera_4",
		"item_i_written",
		"item_k_ash_cell_key",
		"item_k_carson_apartment_key",
		"item_k_chinese_theatre_key",
		"item_k_clinic_cs_key",
		"item_k_clinic_maintenance_key",
		"item_k_clinic_stairs_key",
		"item_k_edane_key",
		"item_k_empire_jezebel_key",
		"item_k_empire_mafia_key",
		"item_k_fu_cell_key",
		"item_k_fu_office_key",
		"item_k_gallery_noir_key",
		"item_k_gimble_key",
		"item_k_hannahs_safe_key",
		"item_k_hitman_ji_key",
		"item_k_hitman_lu_key",
		"item_k_key_generic",
		"item_k_kiki_key",
		"item_k_leopold_int_key",
		"item_k_lilly_key",
		"item_k_lucky_star_murder_key",
		"item_k_malcolm_office_key",
		"item_k_malkavian_refrigerator_key",
		"item_k_murietta_key",
		"item_k_museum_basement_key",
		"item_k_museum_office_key",
		"item_k_museum_storage_key",
		"item_k_museum_storeroom_key",
		"item_k_netcafe_office_key",
		"item_k_oceanhouse_basement_key",
		"item_k_oceanhouse_sewer_key",
		"item_k_oceanhouse_upstairs_key",
		"item_k_oh_front_key",
		"item_k_sarcophagus_key",
		"item_k_shrekhub_four_key",
		"item_k_shrekhub_one_key",
		"item_k_shrekhub_three_key",
		"item_k_skyline_haven_key",
		"item_k_tatoo_parlor_key",
		"item_k_tawni_apartment_key",
		"item_k_tutorial_chopshop_stairs_key",
		"item_m_money_clip",
		"item_m_money_envelope",
		"item_m_wallet",
		"item_p_gargoyle_talisman",
		"item_p_occult_blood_buff",
		"item_p_occult_dexterity",
		"item_p_occult_dodge",
		"item_p_occult_experience",
		"item_p_occult_frenzy",
		"item_p_occult_hacking",
		"item_p_occult_heal_rate",
		"item_p_occult_lockpicking",
		"item_p_occult_obfuscate",
		"item_p_occult_passive_durations",
		"item_p_occult_presence",
		"item_p_occult_regen",
		"item_p_occult_strength",
		"item_p_occult_thaum_damage",
		"item_p_research_hg_computers",
		"item_p_research_hg_dodge",
		"item_p_research_hg_firearms",
		"item_p_research_hg_melee",
		"item_p_research_lg_computers",
		"item_p_research_lg_dodge",
		"item_p_research_lg_firearms",
		"item_p_research_lg_stealth",
		"item_p_research_mg_brawl",
		"item_p_research_mg_finance",
		"item_p_research_mg_melee",
		"item_p_research_mg_security",
		"item_w_avamp_blade",
		"item_w_baseball_bat",
		"item_w_baton",
		"item_w_bush_hook",
		"item_w_chang_blade",
		"item_w_chang_claw",
		"item_w_chang_energy_ball",
		"item_w_chang_ghost",
		"item_w_claws",
		"item_w_claws_ghoul",
		"item_w_claws_protean4",
		"item_w_claws_protean5",
		"item_w_colt_anaconda",
		"item_w_crossbow",
		"item_w_crossbow_flaming",
		"item_w_deserteagle",
		"item_w_fireaxe",
		"item_w_fists",
		"item_w_flamethrower",
		"item_w_gargoyle_fist",
		"item_w_glock_17c",
		"item_w_grenade_frag",
		"item_w_hengeyokai_fist",
		"item_w_ithaca_m_37",
		"item_w_katana",
		"item_w_knife",
		"item_w_mac_10",
		"item_w_manbat_claw",
		"item_w_mingxiao_melee",
		"item_w_mingxiao_spit",
		"item_w_mingxiao_tentacle",
		"item_w_occultblade",
		"item_w_rem_m_700_bach",
		"item_w_remington_m_700",
		"item_w_sabbatleader_attack",
		"item_w_severed_arm",
		"item_w_sheriff_sword",
		"item_w_sledgehammer",
		"item_w_steyr_aug",
		"item_w_supershotgun",
		"item_w_thirtyeight",
		"item_w_tire_iron",
		"item_w_torch",
		"item_w_tzimisce2_claw",
		"item_w_tzimisce2_head",
		"item_w_tzimisce3_claw",
		"item_w_tzimisce_melee",
		"item_w_unarmed",
		"item_w_uzi",
		"item_w_werewolf_attacks",
		"item_w_zombie_fists",
		"keyframe_rope",
		"light",
		"light_dynamic",
		"light_environment",
		"light_spot",
		"momentary_door",
		"momentary_rot_button",
		"move_keyframed",
		"move_rope",
		"mover_keyframe",
		"npc_TestBaseHumanoid",
		"npc_VAndreiBlood",
		"npc_VAnimal",
		"npc_VAsianVampire",
		"npc_VBach",
		"npc_VBatSwarm",
		"npc_VBrujah",
		"npc_VCamera",
		"npc_VCameraSecurity",
		"npc_VChangBros",
		"npc_VChangBrosBlade",
		"npc_VChangBrosClaw",
		"npc_VCombatman",
		"npc_VCop",
		"npc_VDialogPedestrian",
		"npc_VDog",
		"npc_VFrenzyShadow",
		"npc_VGangrel",
		"npc_VGargoyle",
		"npc_VGhoulCroucher",
		"npc_VGuard1",
		"npc_VHengeyokai",
		"npc_VHuman",
		"npc_VHumanCombatPatrol",
		"npc_VHumanCombatant",
		"npc_VHunter",
		"npc_VLasombra",
		"npc_VMalkavian",
		"npc_VManBat",
		"npc_VMercurio",
		"npc_VMingXiao",
		"npc_VMingXiaoTentacle",
		"npc_VMoleman",
		"npc_VNewscaster",
		"npc_VNosferatu",
		"npc_VPedestrian",
		"npc_VPlaceholder",
		"npc_VPlayerController",
		"npc_VProneDialog",
		"npc_VRat",
		"npc_VSabbatGunman",
		"npc_VSabbatLeader",
		"npc_VScurrying",
		"npc_VSheriffMan",
		"npc_VSheriffSwarm",
		"npc_VStalker",
		"npc_VTaxiDriver",
		"npc_VTest",
		"npc_VToreador",
		"npc_VTremere",
		"npc_VTzimisce",
		"npc_VTzimisceHeadClaw",
		"npc_VTzimisceRunner",
		"npc_VVampire",
		"npc_VVampireBoss",
		"npc_VVentrue",
		"npc_VWerewolf",
		"npc_VWolfMorph",
		"npc_VYukie",
		"npc_VZombie",
		"npc_bullseye",
		"npc_crow",
		"npc_generic",
		"npc_generic_bathack",
		"npc_maker",
		"npc_maker_fleshpile",
		"npc_maker_zombie",
		"npc_payphone",
		"npc_sabbat",
		"prop_base",
		"prop_button",
		"prop_clockhand",
		"prop_destructable",
		"prop_doorknob",
		"prop_doorknob_electronic",
		"prop_dynamic",
		"prop_dynamic_ornament",
		"prop_hacking",
		"prop_haunted",
		"prop_keypad",
		"prop_largehull_ignore",
		"prop_model_state",
		"prop_mover",
		"prop_padlock",
		"prop_physics",
		"prop_physics_contested",
		"prop_radio",
		"prop_ragdoll",
		"prop_ragdoll_attached",
		"prop_ragdoll_special",
		"prop_sign",
		"prop_slashable",
		"prop_switch",
		"scripted_sequence",
		"security_camera",
		"simple_physics_brush",
		"simple_physics_prop",
		"trigger",
		"trigger_autosave",
		"trigger_bomb_site",
		"trigger_brush",
		"trigger_changelevel",
		"trigger_checkvolume",
		"trigger_discipline_context",
		"trigger_electric_bugaloo",
		"trigger_environmental_audio",
		"trigger_fog",
		"trigger_hurt",
		"trigger_impact",
		"trigger_inventory_check",
		"trigger_look",
		"trigger_multiple",
		"trigger_once",
		"trigger_player_activity_level",
		"trigger_proximity",
		"trigger_push",
		"trigger_small_hull",
		"trigger_stealth_mod",
		"trigger_teleport",
		"trigger_transition",
		"trigger_werewolf_zone",
		"trigger_wind"
		"window_pane",
		"worldspawn"
	)
	pcCent = pc.GetCenter()
	entListRaw = []
	# player passed class override?
	if len(args) == 1:
		entClassList = (args[0],)
	for entClass in entClassList:
		entsFound = FindClass(entClass)
		for ent in entsFound:
			entListRaw.append((ent, entClass, distanceSquared(pcCent, ent.GetCenter())))
	# filter out entities contained in player
	pcOrigin = pc.GetOrigin()
	entList = []
	for entData in entListRaw:
		if entData[0].GetOrigin() != pcOrigin:
			entList.append(entData)
	del entListRaw
	# sort by distance from player
	entList.sort(lambda a, b: int(a[2] - b[2]))
	# generate output
	print "ENTITIES NEAR PLAYER"
	i = 1
	pzEntList[:] = [] # clear out previous persistent list
	for entData in entList:
		ent = entData[0]
		entClass = entData[1]
		entName = ent.GetName()
		entCent = ent.GetCenter()
		entMdl = ent.GetModelName()
		if (entName == ""):
			entName = "[no targetname]"
		if (entMdl == ""):
			entMdl = "[no model]"
		else:
			entMdl = entMdl[entMdl.rfind("/") + 1:]
		print str(i) + ". " + entClass + ", " + entName + ", " + entMdl + ", \"" + strH(entCent[0]) + " " + strH(entCent[1]) + " " + strH(entCent[2]) + "\""
		# store entity reference and entity class to persistent list
		pzEntList.append((entData[0], entData[1]))
		# max items?
		i = i + 1
		if i > maxItems: break
	return

# znear shortcut
def zn(*args):
	if len(args) == 1:
		znear(args[0])
	else:
		znear()

# ======================================================================
# player functions
# ======================================================================

# set humanity level
def zhum(*args):
	errMsg = "zERR: This function accepts one numeric argument in the range 0-10."
	if len(args) != 1:
		print errMsg
		return
	newH = args[0]
	if (type(newH) != type(1)) or newH < 0 or newH > 10:
		print errMsg
		return
	pc = __main__.FindPlayer()
	pc.HumanityAdd(10)
	pc.HumanityAdd(newH - 10)
	print "HUMANITY LEVEL SET"

# set masquerade violations
def zmasq(*args):
	errMsg = "zERR: This function accepts one numeric argument in the range 0-5."
	if len(args) != 1:
		print errMsg
		return
	newM = args[0]
	if (type(newM) != type(1)) or newM < 0 or newM > 5:
		print errMsg
		return
	pc = __main__.FindPlayer()
	oldM = pc.GetMasqueradeLevel()
	if oldM == newM:
		print "MASQUERADE VIOLATIONS UNCHANGED"
	else:
		pc.ChangeMasqueradeLevel(newM - oldM)
		print "MASQUERADE VIOLATIONS SET"

# max out character stats and inventory
def zbuff():
	pc = __main__.FindPlayer()
	pc.ChangeMasqueradeLevel(-10)
	pc.HumanityAdd(10)
	pc.MoneyAdd(1000000)
	# max all character stats
	statList = (
		"strength",
		"dexterity",
		"stamina",
		"charisma",
		"manipulation",
		"appearance",
		"perception",
		"intelligence",
		"wits",
		"brawl",
		"dodge",
		"intimidation",
		"subterfuge",
		"firearms",
		"melee",
		"security",
		"stealth",
		"computer",
		"finance",
		"investigation",
		"academics"
	)
	for stat in statList:
		pc.BumpStat(stat, 1000)
	# max character disciplines
	discList = (
		"animalism",
		"auspex",
		"celerity",
		"dementation",
		"dominate",
		"fortitude",
		"obfuscate",
		"potence",
		"presence",
		"protean",
		"thaumaturgy"
	)
	for disc in discList:
		if getattr(pc, disc) != -1:
			pc.BumpStat(disc, 10)
	# give all useful items
	itemList = (
		"item_a_hvy_cloth",
		"item_a_lt_leather",
		"item_a_hvy_leather",
		"item_a_body_armor",
		"item_g_bloodpack",
		"item_g_bluebloodpack",
		"item_g_eldervitaepack",
		"item_k_ash_cell_key",
		"item_k_carson_apartment_key",
		"item_k_chinese_theatre_key",
		"item_k_clinic_cs_key",
		"item_k_clinic_maintenance_key",
		"item_k_clinic_stairs_key",
		"item_k_edane_key",
		"item_k_empire_jezebel_key",
		"item_k_empire_mafia_key",
		"item_k_fu_cell_key",
		"item_k_fu_office_key",
		"item_k_gallery_noir_key",
		"item_k_gimble_key",
		"item_k_hannahs_safe_key",
		"item_k_kiki_key",
		"item_k_leopold_int_key",
		"item_k_lilly_key",
		"item_k_lucky_star_murder_key",
		"item_k_malcolm_office_key",
		"item_k_malkavian_refrigerator_key",
		"item_k_murietta_key",
		"item_k_museum_basement_key",
		"item_k_museum_office_key",
		"item_k_museum_storage_key",
		"item_k_museum_storeroom_key",
		"item_k_netcafe_office_key",
		"item_k_oceanhouse_basement_key",
		"item_k_oceanhouse_sewer_key",
		"item_k_oceanhouse_upstairs_key",
		"item_k_oh_front_key",
		"item_k_shrekhub_four_key",
		"item_k_shrekhub_one_key",
		"item_k_shrekhub_three_key",
		"item_k_skyline_haven_key",
		"item_k_tatoo_parlor_key",
		"item_k_tawni_apartment_key",
		"item_k_tutorial_chopshop_stairs_key",
		"item_w_baton",
		"item_w_torch",
		"item_w_fireaxe",
		"item_w_sledgehammer",
		"item_w_bush_hook",
		"item_w_occultblade",
		"item_w_deserteagle",
		"item_w_crossbow_flaming",
		"item_w_flamethrower",
		"item_w_ithaca_m_37",
		"item_w_steyr_aug",
		"item_w_supershotgun",
		"item_w_uzi"
	)
	for item in itemList:
		pc.GiveItem(item)
	zammo()
	print "PLAYER STATS AND INVENTORY MAXED"

# give player some ammo
def zammo():
	pc = __main__.FindPlayer()
	gunList = (
		"item_w_colt_anaconda",
		"item_w_crossbow",
		"item_w_crossbow_flaming",
		"item_w_deserteagle",
		"item_w_flamethrower",
		"item_w_glock_17c",
		"item_w_ithaca_m_37",
		"item_w_mac_10",
		"item_w_steyr_aug",
		"item_w_supershotgun",
		"item_w_thirtyeight",
		"item_w_uzi"
	)
	for gun in gunList:
		pc.GiveAmmo(gun, 999)
	print "PLAYER AMMO FILLED"

# display the player's current angle and position
def zpos():
	pc = __main__.FindPlayer()
	pcCent = pc.GetCenter()
	pcOrig = pc.GetOrigin()
	pcAng = pc.GetAngles()
	print "PLAYER CENTER = " + strH(pcCent[0]) +", " + strH(pcCent[1]) + ", " + strH(pcCent[2])
	print "PLAYER ORIGIN = " + strH(pcOrig[0]) +", " + strH(pcOrig[1]) + ", " + strH(pcOrig[2])
	print "PLAYER ANGLES = " + strH(pcAng[0]) +", " + strH(pcAng[1]) + ", " + strH(pcAng[2])

# ======================================================================
# utility functions
# ======================================================================

# display all ztool commands
def zhelp():
	print "ZTOOLS HELP"
	print "znear() - List entities near to player; zn() for short"
	print "zlist() - Repeat last znear() list"
	print "zgrab() - Grab entity by znear() number or targetname; zg() for short"
	print "zhere()/zthere() - Center on player, and vice-versa"
	print "zbe() - Become grabee"
	print "zorg(x, y, z), zx(n), zy(n), zz(n) - Set X/Y/Z origin"
	print "zix(n), ziy(n), ziz(n) - Increment X/Y/Z origin"
	print "zang(p, w, r), zp(n), zw(n), zr(n) - Set Pitch/yaW/Roll angle"
	print "zip(n), ziw(n), zir(n) - Increment Pitch/yaW/Roll angle"
	print "zhide(), zunhide() - Hide/Unhide grabee"
	print "zmark(), zunmark() - Mark/Unmark grabee"
	print "zinf() - Grabee info"
	print "zammo(), zbuff() - Max ammo/Max everything"
	print "zhum(n)/zmasq(n) - Set Humanity level/Masquerade violations"
	print "ztaxi() - Open all taxi destinations"
	print "zpos() - Player position & angle"
	print "zhate(), zlove(), zkill() - Make all NPCs hate you, love you, and dead, respectively"
	print "zdumpmap(), zupdatemap() - Replicates VPKTool map functionality"
	print "zdumpg() - Dump G to text file"

# display detailed information about entity
def entDump(ent, entClass):
	entName = ent.GetName()
	entCent = ent.GetCenter()
	entOrig = ent.GetOrigin()
	entAng = ent.GetAngles()
	entMdl = ent.GetModelName()
	if (entName == ""):
		entName = "[no targetname]"
	if (entMdl == ""):
		entMdl = "[no model]"
	print "   \"classname\" \"" + entClass + "\""
	print "   \"targetname\" \"" + entName + "\""
	print "   \"model\" \"" + entMdl + "\""
	print "   \"origin\" \"" + str(entOrig[0]) + " " + str(entOrig[1]) + " " + str(entOrig[2]) + "\""
	print "   \"angles\" \"" + str(entAng[0]) + " " + str(entAng[1]) + " " + str(entAng[2]) + "\""
	print "   .GetCenter() = (" + str(entCent[0]) + ", " + str(entCent[1]) + ", " + str(entCent[2]) + ")"

# checks if passed argument list is invalid for entity
# also checks if anything has been grabbed
# this should only be used by functions that operate on grabbed entities
def argsInvalid(args, n):
	if noGrabee(): return 1
	if len(args) != n:
		if n == 0:
			print "zERR: This function does not take any arguments."
		elif n == 1:
			print "zERR: This function takes exactly 1 numeric argument."
		else:
			print "zERR: This function expects exactly " + str(n) + " numeric arguments, separated by commas."
		return 1
	else:
		return 0

# check if there's a current grabbed entity
# used by entity manipulation functions to immediately exit if nothing grabbed
def noGrabee():
	if pzGrabbed == []:
		print "zERR: No entity grabbed. Use zgrab() to grab one."
		return 1
	else:
		return 0

# convert float to string rounded to hundreds decimal place
def strH(n):
	return str(round(n * 100) / 100)

# returns the distance squared between two 3D points
# copied from vamputil.py
def distanceSquared(p1, p2):
    xDistance = (p1[0] - p2[0]) * (p1[0] - p2[0])
    yDistance = (p1[1] - p2[1]) * (p1[1] - p2[1])
    zDistance = (p1[2] - p2[2]) * (p1[2] - p2[2])
    return (xDistance + yDistance + zDistance)

# determine line terminator
def determineTerminator(txt):
	lf = chr(10)
	cr = chr(13)
	if txt.find(cr + lf) > -1: return cr + lf
	if txt.find(lf) > -1: return lf
	if txt.find(cr) > -1: return cr
	return ""

# return current map name
# accepts optional alternate file extension to return
def getMapName(*args):
	map = FindClass("worldspawn")[0]
	m = map.GetModelName()
	mapname = m[m.index("/")+1:m.index(".bsp")]
	if len(args) == 1:
		mapname += "." + str(args[0])
	else:
		mapname += ".bsp"
	return mapname

# ======================================================================
# notify tools successfully loaded
# ======================================================================

print ":::: ZTOOLS LOADED [v0.3] ::::::::::::::::::::::::::::::::::::::::::::::::::"
